from .pddl_parser_types import Token, Expression, Number
from .pddl_tokenizer import PDDLTokenizer
from .domain_parser import DomainParser
from .parsing_utils import parse_signature
from .problem_parser import ProblemParser
from .trajectory_parser import TrajectoryParser



"""Module that contains the tokenization process of the PDDL files."""
import re
from pathlib import Path
from typing import List, Optional

from pddl_plus_parser.lisp_parsers import Token, Expression


class PDDLTokenizer:
    """Class that tokenizes the content of a PDDL file according to the known PDDL scheme."""

    pddl_file_content: List[str]

    def __init__(self, file_path: Optional[Path] = None, pddl_str: Optional[str] = None):
        if file_path is None and pddl_str is None:
            raise ValueError("Cannot receive both the file path and the PDDL str as null.")

        if file_path is not None:
            with open(file_path, "rt", encoding="utf-8") as pddl_file:
                self.pddl_file_content = pddl_file.readlines()

        else:
            self.pddl_file_content = pddl_str.replace("\t", "").split("\n")

    def _is_comment_line(self, line: str) -> bool:
        """Indicates whither or not a line is a comment line

        :param line: the line to test.
        :return: whether or not the line is indeed a comment line.
        """
        return line.strip().startswith(";")

    def tokenize(self) -> List[Token]:
        """Tokenize the PDDL file into tokens."""
        tokens = []
        for line in self.pddl_file_content:
            if self._is_comment_line(line):
                continue

            no_comments_line = re.sub(r";.*", "", line)
            line_tokens = no_comments_line.lower().replace("(", " ( ").replace(")", " ) ").split()
            tokens.extend(line_tokens)

        return tokens

    def read_from_tokens(self, tokens: List[Token]) -> Expression:
        """Extract concrete PDDL expressions from the tokens.

        :param tokens: the list of tokens extracted from the PDDL file.
        :return: concrete PDDL expressions that can be converted to objects.
        """
        if len(tokens) == 0:
            raise SyntaxError("Unexpected EOF")

        token = tokens.pop(0)
        if token == "(":
            expression = []
            while tokens[0] != ")":
                expression.append(self.read_from_tokens(tokens))

            tokens.pop(0)  # pop off ')'
            return expression

        if token == ")":
            raise SyntaxError("Unexpected ) while parsing the expressions")

        return token

    def parse(self) -> Expression:
        """Extracts the expressions from the PDDL file.

        :return: the list of expressions that represent the PDDL file.
        """
        return self.read_from_tokens(self.tokenize())



"""Module test for the numeric trajectory exporter functionality."""

import pytest
from pytest import fixture

from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser
from pddl_plus_parser.models import Domain, State, Problem
from pddl_plus_parser.multi_agent import MultiAgentTrajectoryExporter
from tests.multi_agent_tests.consts import COMBINED_DOMAIN_PATH, COMBINED_PROBLEM_PATH, WOODWORKING_PARSED_PLAN_PATH, \
    WOODWORKING_SHORT_PARSED_PLAN_PATH


@fixture()
def combined_domain() -> Domain:
    domain_parser = DomainParser(COMBINED_DOMAIN_PATH, partial_parsing=False)
    return domain_parser.parse_domain()


@fixture()
def combined_problem(combined_domain: Domain) -> Problem:
    return ProblemParser(problem_path=COMBINED_PROBLEM_PATH, domain=combined_domain).parse_problem()


@fixture()
def multi_agent_trajectory_exporter(combined_domain: Domain) -> MultiAgentTrajectoryExporter:
    return MultiAgentTrajectoryExporter(domain=combined_domain)


def test_create_single_triplet_with_single_state_and_single_call_creates_next_state_correctly(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    test_action_call = "[(nop ),(do-grind grinder0 p0 smooth red varnished colourfragments),(nop ),(nop ),(nop ),(nop ),(nop )]"
    initial_state = State(predicates=combined_problem.initial_state_predicates,
                          fluents=combined_problem.initial_state_fluents)
    result_triplet = multi_agent_trajectory_exporter.create_multi_agent_triplet(initial_state, test_action_call)
    result_next_state = result_triplet.next_state
    grind_cost = initial_state.state_fluents["(grind-cost p0)"].value
    assert result_next_state.state_fluents["(total-cost )"].value == initial_state.state_fluents[
        "(total-cost )"].value + grind_cost


def test_create_single_triplet_with_single_state_and_single_call_with_two_executing_agents_sets_numeric_fluents_correctly(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    test_action_call = "[(nop ),(do-grind grinder0 p0 smooth red varnished colourfragments),(nop ),(nop ),(nop ),(nop ),(nop )]"
    two_agent_action_call = "[(nop ),(nop ),(nop ),(nop ),(do-plane planer0 p2 verysmooth natural varnished),(do-saw-medium saw0 b0 p1 pine rough s3 s2 s1),(nop )]"
    initial_state = State(predicates=combined_problem.initial_state_predicates,
                          fluents=combined_problem.initial_state_fluents)
    result_triplet = multi_agent_trajectory_exporter.create_multi_agent_triplet(initial_state, test_action_call)
    result_next_state = result_triplet.next_state

    result_triplet = multi_agent_trajectory_exporter.create_multi_agent_triplet(result_next_state,
                                                                                two_agent_action_call)
    total_cost_pre_action = 15
    assert result_triplet.next_state.state_fluents["(total-cost )"].value == total_cost_pre_action + 30 + \
           result_triplet.previous_state.state_fluents["(plane-cost p2)"].value


def test_create_single_triplet_with_single_state_and_single_call_with_two_executing_agents_creates_sets_predicates_correctly(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    test_action_call = "[(nop ),(do-grind grinder0 p0 smooth red varnished colourfragments),(nop ),(nop ),(nop ),(nop ),(nop )]"
    two_agent_action_call = "[(nop ),(nop ),(nop ),(nop ),(do-plane planer0 p2 verysmooth natural varnished)," \
                            "(do-saw-medium saw0 b0 p1 pine rough s3 s2 s1),(nop )]"
    initial_state = State(predicates=combined_problem.initial_state_predicates,
                          fluents=combined_problem.initial_state_fluents)
    result_triplet = multi_agent_trajectory_exporter.create_multi_agent_triplet(initial_state, test_action_call)
    result_next_state = result_triplet.next_state

    result_triplet = multi_agent_trajectory_exporter.create_multi_agent_triplet(result_next_state,
                                                                                two_agent_action_call)
    next_state_predicates = result_triplet.next_state.state_predicates
    unused_parts_predicates = [p.untyped_representation for p in next_state_predicates["(unused ?obj)"]]
    available_parts_predicates = [p.untyped_representation for p in next_state_predicates["(available ?obj)"]]
    surface_condition_parts_predicates = [p.untyped_representation for p in
                                          next_state_predicates["(surface-condition ?obj ?surface)"]]
    assert "(unused p1)" not in unused_parts_predicates
    assert "(available p1)" in available_parts_predicates
    assert "(surface-condition p2 smooth)" in surface_condition_parts_predicates


def test_parse_plan_can_parse_complete_plan_content_without_fail(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    try:
        multi_agent_trajectory_exporter.parse_plan(combined_problem, WOODWORKING_PARSED_PLAN_PATH)
    except Exception as e:
        pytest.fail(f"Failed to parse plan: {e}")


def test_parse_plan_can_parse_complete_plan_and_create_correct_length_trajectory(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    triplets = multi_agent_trajectory_exporter.parse_plan(combined_problem, WOODWORKING_PARSED_PLAN_PATH)
    assert len(triplets) == 4


def test_export_returns_readable_list_of_strings_representing_the_trajectory(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    triplets = multi_agent_trajectory_exporter.parse_plan(combined_problem, WOODWORKING_PARSED_PLAN_PATH)
    trajectory_lines = multi_agent_trajectory_exporter.export(triplets)
    for line in trajectory_lines:
        print(line)


def test_export_with_interacting_plan_returns_readable_list_of_strings_representing_the_trajectory(
        multi_agent_trajectory_exporter: MultiAgentTrajectoryExporter, combined_problem: Problem):
    triplets = multi_agent_trajectory_exporter.parse_plan(combined_problem, WOODWORKING_SHORT_PARSED_PLAN_PATH)
    trajectory_lines = multi_agent_trajectory_exporter.export(triplets)
    for line in trajectory_lines:
        print(line)



"""Module that represents a boolean predicate in a PDDL+ model."""
from typing import Dict, Optional, List

from .pddl_type import PDDLType

SignatureType = Dict[str, PDDLType]


class Predicate:
    """Class that represents a boolean predicate."""

    name: str
    signature: SignatureType

    def __init__(self, name: Optional[str] = None, signature: Optional[SignatureType] = None,
                 predicate: Optional["Predicate"] = None):
        if predicate:
            self.name = predicate.name
            self.signature = predicate.signature.copy()

        else:
            self.name = name
            self.signature = signature

    def __eq__(self, other: "Predicate") -> bool:
        """Checks whether or not two predicates are considered equal.

        Equality can be considered if a type inherits from another type as well.

        :param other: the other predicate to compare.
        :return: whether or not the predicates are equal.
        """
        if not self.name == other.name:
            return False

        for parameter_name, parameter_type in self.signature.items():
            if parameter_name not in other.signature:
                return False

            other_param_type = other.signature[parameter_name]
            if not parameter_type.is_sub_type(other_param_type):
                return False

        return True

    @property
    def untyped_representation(self) -> str:
        untyped_signature_str = " ".join(self.signature.keys())
        return f"({self.name} {untyped_signature_str})"

    def __str__(self):
        signature_str_items = []
        for parameter_name, parameter_type in self.signature.items():
            signature_str_items.append(f"{parameter_name} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({self.name} {signature_str})"

    def __hash__(self):
        return hash(self.__str__())


class GroundedPredicate(Predicate):
    """Class defining a grounded predicate."""
    object_mapping: Dict[str, str]

    def __init__(self, name: str, signature: SignatureType, object_mapping: Dict[str, str]):
        super(GroundedPredicate, self).__init__(name=name, signature=signature)
        self.object_mapping = object_mapping

    def __eq__(self, other: "GroundedPredicate") -> bool:
        """Checks whether or not two grounded predicates are considered equal.

        Equality can be considered if a type inherits from another type as well.

        :param other: the other predicate to compare.
        :return: whether or not the predicates are equal.
        """
        if not super(GroundedPredicate, self).__eq__(other):
            return False

        return self.object_mapping == other.object_mapping

    def __ne__(self, other: "GroundedPredicate") -> bool:
        return not self.__eq__(other)

    @property
    def untyped_representation(self) -> str:
        untyped_grounded_signature_str = " ".join(self.object_mapping.values())
        return f"({self.name} {untyped_grounded_signature_str})"

    @property
    def grounded_objects(self) -> List[str]:
        return list(self.object_mapping.values())

    @property
    def lifted_typed_representation(self) -> str:
        return super().__str__()

    @property
    def lifted_untyped_representation(self) -> str:
        return super().untyped_representation

    def __str__(self):
        signature_str_items = []
        for parameter_name, parameter_type in self.signature.items():
            signature_str_items.append(f"{self.object_mapping[parameter_name]} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({self.name} {signature_str})"

    def __hash__(self):
        return hash(self.__str__())



"""Module that represents a PDDL+ domain"""
from typing import List, Dict

from .pddl_action import Action
from .pddl_function import PDDLFunction
from .pddl_object import PDDLConstant
from .pddl_predicate import Predicate
from .pddl_type import PDDLType


class Domain:
    """Class representing the PDDL+ domain object."""

    name: str
    requirements: List[str]
    types: Dict[str, PDDLType]
    constants: Dict[str, PDDLConstant]
    predicates: Dict[str, Predicate]
    functions: Dict[str, PDDLFunction]
    actions: Dict[str, Action]

    # processes: Dict[str, Action] - TBD
    # events: Dict[str, Action] - TBD

    def __init__(self):
        self.actions = {}
        self.constants = {}
        self.functions = {}
        self.types = {}
        self.predicates = {}
        self.requirements = []

    def __str__(self):
        return (
                "< Domain definition: %s\n Requirements: %s\n Predicates: %s\n Functions: %s\n Actions: %s\n "
                "Constants: %s >"
                % (
                    self.name,
                    [req for req in self.requirements],
                    [str(p) for p in self.predicates.values()],
                    [str(f) for f in self.functions.values()],
                    [str(a) for a in self.actions],
                    [str(c) for c in self.constants],
                )
        )



#!/usr/bin/python3
import argparse
import itertools
import random
from pathlib import Path

from pddl_plus_parser.problem_generators import get_problem_template

TEMPLATE_FILE_PATH = Path("counters_template.pddl")


def generate_instance(instance_name: str, num_counters: int, max_int: int) -> str:
    template = get_problem_template(TEMPLATE_FILE_PATH)
    template_mapping = {
        "instance_name": instance_name,
        "domain_name": "fo-counters-rnd",
        "counters_list": " ".join([f"c{i}" for i in range(num_counters)]),
        "counters_initial_values": "\n\t".join(
            [f"(= (value c{i}) {random.randint(0, max_int)})" for i in range(num_counters)]),
        "counters_rate_values": "\n\t".join([f"(= (rate_value c{i}) 0)" for i in range(num_counters)]),
        "max_int_value": max_int,
        "counters_final_values": "\n\t".join(
            [f"(<= (+ (value c{i}) 1) (value c{i + 1}))" for i in range(num_counters - 1)])
    }
    return template.substitute(template_mapping)


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Generate sailing planning instance")
    parser.add_argument("--num_counters", required=True, help="number of counters in the problems")
    parser.add_argument("--max_int", required=True, help="Maximal integer value")
    args = parser.parse_args()
    return args


def generate_multiple_problems(
        min_counters: int, max_counters: int, max_int: int, output_folder: Path):
    counters_range = [i for i in range(min_counters, max_counters + 1)]
    for num_counters in counters_range:
        with open(output_folder / f"pfile{num_counters}_{random.randint(0, 100)}.pddl", "wt") as problem_file:
            problem_file.write(
                generate_instance(f"instance_{num_counters}_{max_int}", num_counters, max_int))


def main():
    generate_multiple_problems(2, 40, 100, Path("C:\Argaman\Planning\Minecraft\more_domains\counters"))


if __name__ == '__main__':
    main()



import enum
import random
import sys


class typeStatus(enum.Enum):
    OFF = 0
    ON = 1


typing = typeStatus


def rnd(limit):
    return 1 + int(limit * random.random())


def rnd2(limit):
    return limit / 2 + rnd(limit / 2)


def rnd3(locs, limit):
    return int(locs * (1.0 + ((limit - 1) * random.random())))


class locatable:
    def __init__(self, locs=-1):
        if locs == -1:
            self.location = 0
        else:
            self.location = rnd(locs)


class Truck(locatable):
    def __init__(self, locs, maxc):
        locatable.__init__(self, locs)
        self.capacity = rnd2(maxc)
        self.speed = rnd(10)


class Pallet(locatable):
    def __init__(self, locs):
        locatable.__init__(self, locs)
        self.topcrate = 0
        self.dtopcrate = 0

maxWeight = 0

class Crate(locatable):

    def __init__(self, ps, maxw):
        locatable.__init__(self)
        self.pallet = rnd(ps)
        self.weight = rnd(maxw)
        self.dpallet = 0
        self.dsurface = 0
        self.surface = int()
        global maxWeight
        maxWeight = self.weight if self.weight > maxWeight else maxWeight

class Hoist(locatable):
    def __init__(self, locs):
        locatable.__init__(self, locs)
        self.powr = rnd(10)


class Map:
    def __init__(self, locs):
        self.distances = []
        for i in range(locs):
            self.distances.append(list())
            for j in range(locs):
                self.distances[i].append(rnd(10))


class DepotDescriptor:
    class ProblemType(enum.Enum):
        STRIPS = 0
        NUMERIC = 1
        SIMPLETIMED = 2
        TIMED = 3

    def __init__(self, dps, dsts, trks, plts, hsts, crts, tp=ProblemType.STRIPS, mxwt=0, mxcp=0):
        self.numDepots = dps
        self.numDistributors = dsts
        self.numTrucks = trks
        self.numPallets = plts
        self.numHoists = hsts
        self.numCrates = crts
        self.probtype = tp
        self.maxCapacity = mxcp
        self.maxWeight = mxwt


class Depot:
    def __init__(self, s, d, f):
        self.__seed = s
        self.__probtype = d.probtype
        random.seed(s)
        self.__numDepots = d.numDepots
        self.__numDistributors = d.numDistributors
        self.numTrucks = d.numTrucks
        self.locs = self.__numDepots + self.__numDistributors
        self.__m = Map(self.locs)
        self.numPallets = max(d.numPallets, self.locs)
        self.numHoists = max(d.numHoists, self.locs)
        self.__trucks = []
        self.__pallets = []
        self.__hoists = []
        self.__crates = []
        self.file = f
        for i in range(self.numTrucks):
            t = Truck(self.locs, d.maxCapacity)
            self.__trucks.append(t)
        for i in range(self.numHoists):
            h = Hoist(self.locs)
            if i < self.locs:
                h.location = i + 1
            self.__hoists.append(h)
        for i in range(self.numPallets):
            p = Pallet(self.locs)
            if i < self.locs:
                p.location = i + 1
            self.__pallets.append(p)
        for i in range(d.numCrates):
            c = Crate(self.numPallets, d.maxWeight)
            c.location = self.__pallets[c.pallet - 1].location
            c.surface = self.__pallets[c.pallet - 1].topcrate
            self.__pallets[c.pallet - 1].topcrate = i + 1
            self.__crates.append(c)
        for i in range(2 * d.numCrates):
            c = rnd(d.numCrates) - 1
            if self.__crates[c].dpallet:
                continue
            self.__crates[c].dpallet = rnd(self.numPallets)
            self.__crates[c].dsurface = self.__pallets[self.__crates[c].dpallet - 1].dtopcrate
            self.__pallets[self.__crates[c].dpallet - 1].dtopcrate = c + 1

    def __location(self, i):
        if type(i) != int:
            i = i.location - 1
        if i < self.__numDepots:
            self.file.write(f"depot{i}")
        else:
            self.file.write(f"distributor{i - self.__numDepots}")


    def __del__(self):
        del self.__m

    def write(self):  
        global maxWeight
        self.file.write(
            f"(define (problem depotprob{self.__seed}) (:domain Depot)\n(:objects\n\t")
        for i in range(self.__numDepots):
            self.file.write(f"depot{i} ")
        if typing == typeStatus.ON:
            self.file.write("- Depot\n\t")
        for i in range(self.__numDistributors):
            self.file.write(f"distributor{i} ")
        if typing == typeStatus.ON:
            self.file.write("- Distributor\n\t")
        for i in range(len(self.__trucks)):
            self.file.write(f"truck{i} ")
        if typing == typeStatus.ON:
            self.file.write("- Truck\n\t")
        for i in range(len(self.__pallets)):
            self.file.write(f"pallet{i} ")
        if typing == typeStatus.ON:
            self.file.write("- Pallet\n\t")
        for i in range(len(self.__crates)):
            self.file.write(f"crate{i} ")
        if typing == typeStatus.ON:
            self.file.write("- Crate\n\t")
        for i in range(len(self.__hoists)):
            self.file.write(f"hoist{i} ")
        if typing == typeStatus.ON:
            self.file.write("- Hoist")
        self.file.write(")\n(:init\n")
        for i in range(len(self.__pallets)):
            if typing == typeStatus.OFF:
                self.file.write(f"\t(pallet pallet{i})\n")
                self.file.write(f"\t(surface pallet{i})\n")
            self.file.write(f"\t(at pallet{i} ")
            self.__location(self.__pallets[i])
            self.file.write(")\n\t(clear ")
            if self.__pallets[i].topcrate:
                self.file.write(
                    f"crate{self.__pallets[i].topcrate - 1})\n")
            else:
                self.file.write(f"pallet{i})\n")
        for i in range(len(self.__trucks)):
            if typing == typeStatus.OFF:
                self.file.write(f"\t(truck truck{i})\n")
            self.file.write(f"\t(at truck{i} ")
            self.__location(self.__trucks[i])
            self.file.write(")\n")
            if self.__probtype == DepotDescriptor.ProblemType.TIMED:
                self.file.write(
                    f"\t(= (speed truck{i}) {self.__trucks[i].speed})\n")
            if self.__probtype == DepotDescriptor.ProblemType.NUMERIC:
                self.file.write(
                    f"\t(= (current_load truck{i}) 0)\n\t(= (load_limit truck{i}) ")
                if ((2 * self.__trucks[i].capacity) < maxWeight) or i == 0:
                    self.file.write(
                        f"{str(self.__trucks[i].capacity + maxWeight)})\n")
                else:
                    self.file.write(
                        f"{str(self.__trucks[i].capacity)})\n")
        for i in range(len(self.__hoists)):
            if typing == typeStatus.OFF:
                self.file.write(f"\t(hoist hoist{i})\n")
            self.file.write(f"\t(at hoist{i} ")
            self.__location(self.__hoists[i])
            self.file.write(f")\n\t(available hoist{i})\n")
            if self.__probtype == DepotDescriptor.ProblemType.TIMED:
                self.file.write(
                    f"\t(= (power hoist{i}) {self.__hoists[i].powr})\n")
        for i in range(len(self.__crates)):
            if typing == typeStatus.OFF:
                self.file.write(f"\t(crate crate{i})\n")
                self.file.write(f"\t(surface crate{i})\n")
            self.file.write(f"\t(at crate{i} ")
            self.__location(self.__crates[i])
            self.file.write(f")\n\t(on crate{i} ")
            if self.__crates[i].surface:
                self.file.write(
                    f"crate{self.__crates[i].surface - 1})\n")
            else:
                self.file.write(
                    f"pallet{self.__crates[i].pallet - 1})\n")
            if self.__probtype == DepotDescriptor.ProblemType.NUMERIC or self.__probtype == DepotDescriptor.ProblemType.TIMED:
                self.file.write(
                    f"\t(= (weight crate{i}) {self.__crates[i].weight})\n")
        if typing == typeStatus.OFF:
            for i in range(self.__numDepots + self.__numDistributors):
                self.file.write("\t(place ")
                self.__location(i)
                self.file.write(")\n")
        if self.__probtype == DepotDescriptor.ProblemType.TIMED:
            for i in range(len(self.__m.distances)):
                for j in range(len(self.__m.distances[i])):
                    self.file.write("\t(= (distance ")
                    self.__location(i)
                    self.file.write(" ")
                    self.__location(j)
                    self.file.write(") ")
                    if j != i:
                        self.file.write(
                            f"{self.__m.distances[i][j]})\n")
                    else:
                        self.file.write("0)\n")

        if self.__probtype == DepotDescriptor.ProblemType.NUMERIC:
            self.file.write("\t(= (fuel-cost) 0)\n")

        self.file.write(")\n\n(:goal (and\n")
        for i in range(len(self.__crates)):
            if self.__crates[i].dpallet:
                if self.__crates[i].dsurface:
                    self.file.write(
                        f"\t\t(on crate{i} crate{self.__crates[i].dsurface - 1})\n")
                else:
                    self.file.write(
                        f"\t\t(on crate{i} pallet{self.__crates[i].dpallet - 1})\n")
        self.file.write("\t)\n)")
        if self.__probtype != DepotDescriptor.ProblemType.STRIPS and self.__probtype != DepotDescriptor.ProblemType.NUMERIC:
            self.file.write("\n\n(:metric minimize (total-time))")
        if self.__probtype == DepotDescriptor.ProblemType.NUMERIC:
            if rnd(10) < 4:
                self.file.write("\n\n(:metric minimize (fuel-cost))")
            else:
                self.file.write("\n\n(:metric minimize (total-time))")
        self.file.write(")\n")


def usage():
    print("Usage: gdep <seed> [-s|-t|-n|-w <weight>|-c <capacity>|-f <filename>]\n\t\t<#depots> <#distributors> <#trucks> <#pallets> <#hoists> <#crates>\n\n\tOptions:\n\tu: untyped\n\ts: simple-time\n\tt: time\n\tn: numeric\n\tw: provide maximum weight for crates\n\tc: maximum capacity for trucks\n\tf: optional file for output\n\n\tAll numbers are integers.\n\n")
    exit(0)


def commandLine(seed, filename, argc, argv):
    probtype = DepotDescriptor.ProblemType.STRIPS
    global typing
    wgt = 1
    cap = 1
    nxt = 0
    val = [0 for _ in range(6)]

    if argc <= 0:
        usage()

    seed = int(argv[0])
    argc -= 1
    argv = argv[1:]

    while argc > 0:
        if argv[0][0] == '-':
            o = argv[0][1]
            if o == 't':
                probtype = DepotDescriptor.ProblemType.TIMED
            elif o == 's':
                probtype = DepotDescriptor.ProblemType.SIMPLETIMED
            elif o == 'n':
                probtype = DepotDescriptor.ProblemType.NUMERIC
            elif o == 'u':
                typing = typeStatus.OFF
            else:
                argc -= 1
                argv = argv[1:]
                if argc < 0:
                    usage()
                if o == 'w':
                    wgt = int(argv[0])
                elif o == 'c':
                    cap = int(argv[0])
                elif o == 'f':
                    filename = argv[0]
                else:
                    usage()
            argc -= 1
            argv = argv[1:]
        else:
            if nxt == 6:
                usage()
            val[nxt] = int(argv[0])
            nxt += 1
            argv = argv[1:]
            argc -= 1
    if nxt < 6:
        usage()
    return DepotDescriptor(val[0], val[1], val[2], val[3], val[4], val[5], probtype, wgt, cap), filename, seed


def main(argc, argv):
    global typing
    typing = typeStatus.ON
    seed = int()
    filename = ''
    d, filename, seed = commandLine(seed, filename, argc - 1, argv[1:])
    if filename != '':
        f = open(filename, 'w')
        dp = Depot(seed, d, f)
        dp.write()
        f.close()


if __name__ == '__main__':
    argv = sys.argv
    argc = len(argv)
    main(argc, argv)



"""Module test for the multi-agent to single agent domain convertion."""
from pytest import fixture

from pddl_plus_parser.multi_agent import MultiAgentDomainsConverter
from tests.multi_agent_tests.consts import MULTI_AGENT_DATA_DIRECTORY


@fixture()
def domain_converter() -> MultiAgentDomainsConverter:
    return MultiAgentDomainsConverter(MULTI_AGENT_DATA_DIRECTORY)


def test_locate_domains_returns_domain_with_correct_name(domain_converter: MultiAgentDomainsConverter):
    """Test that the locate_domains method returns a domain with a correct name."""
    domain = domain_converter.locate_domains()
    assert domain.name == "woodworking"


def test_locate_domains_returns_all_private_and_public_predicates(domain_converter: MultiAgentDomainsConverter):
    """Test that the locate_domains method returns a domain with all private and public predicates."""
    domain = domain_converter.locate_domains()
    assert "grind-treatment-change" in domain.predicates
    assert "boardsize-successor" in domain.predicates
    assert len(domain.predicates) == 14


def test_locate_domains_returns_agents_actions(domain_converter: MultiAgentDomainsConverter):
    """Test that the locate_domains method returns a domain with all agents' actions."""
    domain = domain_converter.locate_domains()
    assert "do-grind" in domain.actions
    assert "do-saw-small" in domain.actions
    assert "do-saw-medium" in domain.actions
    assert "do-saw-large" in domain.actions



"""Module that encapsulates the numeric trajectory functionalities."""
import logging
from pathlib import Path
from typing import List, NoReturn, Optional

from pddl_plus_parser.models import Domain, Problem, Operator, State, ActionCall


def parse_action_call(action_call: str) -> ActionCall:
    """Parses the string representing the action call in the plan sequence.

    :param action_call: the string representing the action call.
    :return: the object representing the action name and its parameters.
    """
    action_data = action_call.lower().replace("(", " ( ").replace(")", " ) ").split()
    action_data = action_data[1:-1]
    return ActionCall(name=action_data[0], grounded_parameters=action_data[1:])


class TrajectoryTriplet:
    """Class representing a single trajectory triplet."""
    previous_state: State
    operator: Operator
    next_state: State

    def __init__(self, previous_state: State, op: Operator, next_state: State):
        self.previous_state = previous_state
        self.operator = op
        self.next_state = next_state

    def __str__(self):
        return f"previous state: {self.previous_state.serialize()}\n" \
               f"operator: {str(self.operator)}\n" \
               f"next state: {self.next_state.serialize()}"


class TrajectoryExporter:
    """Export trajectories in the appropriate format."""

    domain: Domain
    logger: logging.Logger

    def __init__(self, domain: Domain):
        self.domain = domain
        self.logger = logging.getLogger(__name__)

    def _read_plan(self, plan_file_path: Path) -> List[str]:
        """Read the plan file and exports the lines with the actions.

        :param plan_path: the path to the plan file.
        :return: the action sequence.
        """
        self.logger.debug(f"Reading the plan in the path {plan_file_path}")
        with open(plan_file_path, "rt") as plan_file:
            return plan_file.readlines()

    def create_single_triplet(self, previous_state: State, action_call: str) -> TrajectoryTriplet:
        """Create a single trajectory triplet by applying the action on the input state.

        :param previous_state: the state that the action is being applied on.
        :param action_call: the string representation of the grounded action call.
        :return: the new triplet containing (s,a,s').
        """
        self.logger.info(f"Trying to apply the action - {action_call} on the state - {previous_state.serialize()}")
        action_descriptor = parse_action_call(action_call)
        operator = Operator(action=self.domain.actions[action_descriptor.name],
                            domain=self.domain,
                            grounded_action_call=action_descriptor.parameters)
        next_state = operator.apply(previous_state)
        return TrajectoryTriplet(previous_state=previous_state,
                                 op=operator,
                                 next_state=next_state)

    def parse_plan(self, problem: Problem, plan_path: Optional[Path] = None,
                   action_sequence: Optional[List[str]] = None) -> List[TrajectoryTriplet]:
        """Parse the input plan file to create the trajectory.

        :return: the list of triplets that was generated using the plan.
        """
        self.logger.info("Parsing the plan to extract the grounded operators.")
        plan_actions = action_sequence if action_sequence is not None else self._read_plan(plan_path)
        initial_state_predicates = problem.initial_state_predicates
        initial_state_numeric_fluents = problem.initial_state_fluents
        previous_state = State(predicates=initial_state_predicates, fluents=initial_state_numeric_fluents, is_init=True)
        triplets = []
        self.logger.debug("Starting to create the trajectory triplets.")
        for grounded_action_call in plan_actions:
            triplet = self.create_single_triplet(previous_state, grounded_action_call)
            triplets.append(triplet)
            previous_state = triplet.next_state

        return triplets

    @staticmethod
    def export(triplets: List[TrajectoryTriplet]) -> List[str]:
        """Export the input triplets as a valid trajectory object.

        :param triplets: the objects representing the triplets generated from the plan sequence.
        :return: a list of strings representing the trajectory.
        """
        serialized_trajectory = []
        first_state = triplets[0].previous_state
        serialized_trajectory.append(first_state.serialize())
        for triplet in triplets:
            serialized_trajectory.append(f"(operator: {str(triplet.operator)})\n")
            serialized_trajectory.append(triplet.next_state.serialize())

        serialized_trajectory[0] = f"({serialized_trajectory[0]}"
        serialized_trajectory[-1] = f"{serialized_trajectory[-1]})"
        return serialized_trajectory

    def export_to_file(self, triplets: List[TrajectoryTriplet], output_path: Path) -> NoReturn:
        """Export the trajectory to a file.

        :param triplets: the trajectory triples.
        :param output_path: the path to the output file.
        """
        trajectory_lines = self.export(triplets)
        with open(output_path, "wt") as output_path:
            output_path.writelines(trajectory_lines)



from pddl_plus_parser.lisp_parsers import PDDLTokenizer
from tests.lisp_parsers_tests.consts import TEST_NUMERIC_PROBLEM


def test_simple_parsing():
    test_program = "(begin (define r 10) (* pi (* r r)))"
    simple_lisp_parser = PDDLTokenizer(pddl_str=test_program)
    tokens = simple_lisp_parser.parse()
    assert tokens == ['begin', ['define', 'r', '10'], ['*', 'pi', ['*', 'r', 'r']]]


def test_simple_predicate_tokenization():
    test_predicate = "(available ?obj - woodobj)"
    simple_lisp_parser = PDDLTokenizer(pddl_str=test_predicate)
    tokens = simple_lisp_parser.parse()
    assert tokens == ['available', '?obj', '-', 'woodobj']


def test_multiple_predicates_tokenization():
    test_predicates = """((available ?obj - woodobj)
    (surface-condition ?obj - woodobj ?surface - surface))"""
    multi_predicate_tokenizer = PDDLTokenizer(pddl_str=test_predicates)
    tokens = multi_predicate_tokenizer.parse()
    assert tokens == [['available', '?obj', '-', 'woodobj'],
                      ['surface-condition', '?obj', '-', 'woodobj', '?surface', '-', 'surface']]


def test_parse_action_yields_correct_action_ast():
    test_action_str = """(do-spray-varnish
    	:parameters   (?m - spray-varnisher ?x - part ?newcolour - acolour ?surface - surface)
    	:precondition (and (available ?x) (has-colour ?m ?newcolour))
    	:effect       (and (treatment ?x varnished) (colour ?x ?newcolour)))"""
    multi_predicate_tokenizer = PDDLTokenizer(pddl_str=test_action_str)
    tokens = multi_predicate_tokenizer.parse()
    expected_tokens = ['do-spray-varnish',
                       ':parameters',
                       ['?m', '-', 'spray-varnisher', '?x', '-', 'part', '?newcolour', '-', 'acolour', '?surface', '-',
                        'surface'],
                       ':precondition', ['and', ['available', '?x'], ['has-colour', '?m', '?newcolour']],
                       ':effect', ['and', ['treatment', '?x', 'varnished'], ['colour', '?x', '?newcolour']]]
    assert tokens == expected_tokens

def test_parse_problem():
    test_problem_tokenizer = PDDLTokenizer(file_path=TEST_NUMERIC_PROBLEM)
    print(test_problem_tokenizer.parse())



"""Module test for the multi-agent to single agent problem convertion."""
from pytest import fixture

from pddl_plus_parser.multi_agent import MultiAgentProblemsConverter
from tests.multi_agent_tests.consts import MULTI_AGENT_DATA_DIRECTORY, COMBINED_DOMAIN_PATH


@fixture()
def problem_converter() -> MultiAgentProblemsConverter:
    return MultiAgentProblemsConverter(MULTI_AGENT_DATA_DIRECTORY, problem_file_prefix="problem")


def test_combine_problems_returns_problem_with_correct_name(problem_converter: MultiAgentProblemsConverter):
    combined_problem = problem_converter.combine_problems(COMBINED_DOMAIN_PATH)
    assert combined_problem.name == "wood-prob"


def test_combine_problems_returns_problem_with_all_private_and_public_objects(
        problem_converter: MultiAgentProblemsConverter):
    combined_problem = problem_converter.combine_problems(COMBINED_DOMAIN_PATH)
    assert "highspeed-saw0" in combined_problem.objects


def test__combine_problems_returns_problem_with_initial_state_with_private_objects(
        problem_converter: MultiAgentProblemsConverter):
    combined_problem = problem_converter.combine_problems(COMBINED_DOMAIN_PATH)
    assert "(empty ?agent)" in combined_problem.initial_state_predicates
    empty_predicates = combined_problem.initial_state_predicates["(empty ?agent)"]
    assert "(empty highspeed-saw0)" in [p.untyped_representation for p in empty_predicates]



"""Module that contains the parser for PDDL+ domain files."""
import logging
import types
from pathlib import Path
from typing import List, Dict, Union, NoReturn

from pddl_plus_parser.models import Domain, PDDLType, Predicate, PDDLConstant, PDDLFunction, Action, SignatureType, \
    NumericalExpressionTree, construct_expression_tree
from .parsing_utils import parse_signature
from .pddl_tokenizer import PDDLTokenizer

ObjectType = PDDLType(name="object", parent=None)

COMPARISON_OPS = ["=", "<=", ">=", ">", "<"]
ASSIGNMENT_OPS = ["assign", "increase", "decrease"]
EQUALITY_OPERATOR = "="


class DomainParser:
    """Class that parses PDDL+ domain files."""

    tokenizer: PDDLTokenizer
    logger: logging.Logger
    partial_parsing: bool
    enable_conjunctions: bool

    def __init__(self, domain_path: Path, partial_parsing: bool = False, enable_conjunctions: bool = False):
        self.tokenizer = PDDLTokenizer(domain_path)
        self.logger = logging.getLogger(__name__)
        self.partial_parsing = partial_parsing
        self.enable_conjunctions = enable_conjunctions

    def parse_types(self, types: List[str]) -> Dict[str, PDDLType]:
        """Parse the types of the domain.

        :param types: the list containing the
        :return: a mapping between the type name and the appropriate PDDLType object.
        """
        self.logger.info("Starting to parse the types in the domain!")
        pddl_types = {}
        same_types_objects = []
        index = 0
        while index < len(types):
            if types[index] != "-":
                same_types_objects.append(types[index])
                index += 1
                continue

            pddl_type = types[index + 1]
            parent_type = pddl_types.get(pddl_type, PDDLType(name=pddl_type, parent=ObjectType))
            pddl_types.update({descendant_typ_name: PDDLType(name=descendant_typ_name, parent=parent_type)
                               for descendant_typ_name in same_types_objects})
            same_types_objects = []
            index += 2
            continue

        if len(same_types_objects) > 0:
            pddl_types.update({type_name: PDDLType(name=type_name, parent=ObjectType)
                               for type_name in same_types_objects})

        pddl_types["object"] = ObjectType
        self.logger.debug(f"Extracted {len(pddl_types)} types while parsing the types AST.")
        return pddl_types

    def parse_constants(self, constants_ast: List[str], domain_types: Dict[str, PDDLType]) -> Dict[str, PDDLConstant]:
        """Parses the constants that appear in the constants part of the domain AST.

        :param constants_ast: the constants that were parsed in the domain AST.
        :param domain_types: the types that exist in the domain.
        :return: a list containing all typed constants.
        """
        self.logger.info("Starting to parse the constants of the domain!")
        constants = {}
        same_type_constants = []
        type_marker_reached = False
        for constant_name in constants_ast:
            if type_marker_reached:
                # This means that the name of the constant here is actually its type.
                if constant_name not in domain_types:
                    raise SyntaxError("Received invalid type for the constants!")

                constants.update(
                    {name: PDDLConstant(name, domain_types[constant_name]) for name in same_type_constants})
                type_marker_reached = False
                same_type_constants = []
                continue

            if constant_name == "-":
                type_marker_reached = True
                continue

            same_type_constants.append(constant_name)

        self.logger.debug(f"Extracted {len(constants)} from the domain.")
        return constants

    def parse_predicate(self, predicate: List[str], domain_types: Dict[str, PDDLType]) -> Predicate:
        """Parse a single predicate from an AST representation in the PDDL domain.

        :param predicates_ast: the predicate in the form of an AST list of strings.
        :param domain_types: the types that were extracted from the domain.
        :return: the predicate object that represents the list of strings that were given.
        """
        self.logger.info(f"Parsing the predicate represented by the AST - {predicate}")
        predicate_name = predicate[0]
        if (len(predicate[1:]) % 3) != 0:
            raise SyntaxError(f"Received a predicate with a wrong signature - {predicate[1:]}")

        signature_items = iter(predicate[1:])
        predicate_signature = parse_signature(signature_items, domain_types)
        extracted_predicate = Predicate(name=predicate_name, signature=predicate_signature)
        self.logger.debug(f"Finished extracting the predicate - {extracted_predicate}")
        return extracted_predicate

    def parse_untyped_predicate(self, untyped_predicate: List[str], action_signature: SignatureType,
                                domain_constants: Dict[str, PDDLConstant] = {}) -> Predicate:
        """Parse an untyped predicate that appears in actions.

        :param untyped_predicate: the untyped predicate that needs to be matched to the typed predicate from the
            predicates definitions.
        :param action_signature: the signed signature of the action.
        :return: the predicate including the relevant types.
        :param domain_constants: the constants that are defined in the domain.
        """
        self.logger.info(f"Parsing the untyped action predicate represented by the AST - {untyped_predicate}")
        predicate_name = untyped_predicate[0]
        possible_signed_objects = {key: val for key, val in action_signature.items()}
        possible_signed_objects.update({const_name: const.type for const_name, const in domain_constants.items()})
        # Since we assume that the order in maintained in the predicates we can match the signatures.
        signed_signature = {parameter_name: possible_signed_objects[parameter_name] for
                            parameter_name in untyped_predicate[1:]}

        signed_action_predicate = Predicate(name=predicate_name, signature=signed_signature)
        self.logger.debug(f"Extracted the predicate - {signed_action_predicate}")
        return signed_action_predicate

    def parse_predicates(self, predicates_ast: List[List[str]],
                         domain_types: Dict[str, PDDLType]) -> Dict[str, Predicate]:
        """Parses the predicates that appear in the predicates parsed AST.

        :param predicates_ast: the AST that contains the predicates of the domain.
        :param domain_types: the types that exist in the domain.
        :return: a mapping between the predicate name and the predicate itself.
        """
        predicates = {}
        for predicate in predicates_ast:
            if predicate[0] == ":private":
                for private_predicate in predicate[1:]:
                    extracted_private_predicate = self.parse_predicate(private_predicate, domain_types)
                    predicates[extracted_private_predicate.name] = extracted_private_predicate

                continue

            extracted_predicate = self.parse_predicate(predicate, domain_types)
            predicates[extracted_predicate.name] = extracted_predicate

        return predicates

    def parse_functions(self, functions_ast: List[List[str]],
                        domain_types: Dict[str, PDDLType]) -> Dict[str, PDDLFunction]:
        """Parses the functions that appear in the functions parsed AST.

        :param functions_ast: the AST that contains the numerical functions of the domain.
        :param domain_types: the types that exist in the domain.
        :return: a mapping between a function name and the function itself.
        """
        self.logger.info("Starting to parse the function' definition in the domain data.")
        functions = {}
        for function_items in functions_ast:
            function_name = function_items[0]
            if (len(function_items[1:]) % 3) != 0:
                raise SyntaxError(f"Received a function with a wrong signature - {function_items[1:]}")

            signature_items = iter(function_items[1:])
            function_signature = parse_signature(signature_items, domain_types)
            functions[function_name] = PDDLFunction(name=function_name, signature=function_signature)

        return functions

    def parse_disjunctive_numeric_preconditions(
            self, numeric_preconditions_node: List[Union[str, List[str]]],
            new_action: Action, domain_functions: Dict[str, PDDLFunction]) -> NoReturn:
        """Parse a set of disjunctive numeric preconditions.

        :param numeric_preconditions_node: the node that contains the numeric preconditions.
        :param new_action: the action that is currently being parsed.
        :param domain_functions: the functions that are defined in the domain.
        """
        self.logger.info("Starting to parse the disjunctive numeric preconditions!")
        for conditions_set_ast in numeric_preconditions_node:
            if conditions_set_ast[0] != "and":
                raise SyntaxError(
                    f"Only accepting conjunctive preconditions! Action - {new_action.name} does not conform!")

            conditions_set = {NumericalExpressionTree(construct_expression_tree(node, domain_functions)) for node in
                              conditions_set_ast[1:]}
            new_action.disjunctive_numeric_preconditions.append(conditions_set)

    def parse_preconditions(self, preconditions_ast: List[Union[str, List[str]]], new_action: Action,
                            domain_functions: Dict[str, PDDLFunction],
                            domain_predicates: Dict[str, Predicate],
                            domain_constants: Dict[str, PDDLConstant]) -> NoReturn:
        """Parse the preconditions of a single action.

        :param preconditions_ast: the AST representation of the action's preconditions.
        :param new_action: the action that is currently being parsed.
        :param domain_functions: the functions that exist in the domain.
        :param domain_predicates: the predicates that exist in the domain.
        :param domain_constants: the constants that might exist in the domain.
        """
        new_action.positive_preconditions = set()
        new_action.negative_preconditions = set()
        new_action.equality_preconditions = set()
        new_action.inequality_preconditions = set()
        new_action.numeric_preconditions = set()
        if len(preconditions_ast) == 0:
            self.logger.warning("Received an action with no preconditions.")
            return

        if preconditions_ast[0] != "and" and not self.enable_conjunctions:
            raise SyntaxError(f"Only accepting conjunctive preconditions! Action - {new_action.name} does not conform!")

        for precondition_node in preconditions_ast[1:]:
            if precondition_node[0] in domain_predicates:
                new_action.positive_preconditions.add(
                    self.parse_untyped_predicate(precondition_node, new_action.signature, domain_constants))
                continue

            if precondition_node[0] == "not":
                inner_node = precondition_node[1]
                if inner_node[0] == EQUALITY_OPERATOR:
                    self.logger.debug("Adding new lifted objects that should be tested for inequality")
                    new_action.inequality_preconditions.add((inner_node[1], inner_node[2]))
                    continue

                new_action.negative_preconditions.add(
                    self.parse_untyped_predicate(precondition_node[1], new_action.signature, domain_constants))
                continue

            if precondition_node[0] == EQUALITY_OPERATOR:
                if not isinstance(precondition_node[1], List):
                    self.logger.debug("Adding new lifted objects that should be tested for equality")
                    new_action.equality_preconditions.add((precondition_node[1], precondition_node[2]))
                    continue

            if precondition_node[0] == "or":
                self.logger.debug("Assuming that the OR operator is used for the disjunction of numeric preconditions.")
                self.parse_disjunctive_numeric_preconditions(precondition_node[1:], new_action, domain_functions)

            if precondition_node[0] in COMPARISON_OPS:
                numerical_precondition = NumericalExpressionTree(
                    construct_expression_tree(precondition_node, domain_functions))
                new_action.numeric_preconditions.add(numerical_precondition)
                continue

    def parse_effects(self, effects_ast: List[Union[str, List[str]]], new_action: Action,
                      domain_functions: Dict[str, PDDLFunction],
                      domain_predicates: Dict[str, Predicate],
                      domain_constants: Dict[str, PDDLConstant]) -> NoReturn:
        """Parse the effects of a single action.

        :param effects_ast: the AST representation of the action's effects.
        :param new_action: the action that is currently being parsed.
        :param domain_functions: the functions that exist in the domain.
        :param domain_predicates: the predicates that are defined in the domain.
        :param domain_constants: the domains that might exist in the domain.
        """
        new_action.add_effects = set()
        new_action.delete_effects = set()
        new_action.numeric_effects = set()
        if effects_ast[0] != "and":
            raise SyntaxError(f"Only accepting conjunctive effects! Action - {new_action.name} does not conform!")

        for effect_node in effects_ast[1:]:
            if effect_node[0] in domain_predicates:
                new_action.add_effects.add(
                    self.parse_untyped_predicate(effect_node, new_action.signature, domain_constants))
                continue

            if effect_node[0] == "not":
                new_action.delete_effects.add(
                    self.parse_untyped_predicate(effect_node[1], new_action.signature, domain_constants))
                continue

            if effect_node[0] in ASSIGNMENT_OPS:
                numerical_precondition = NumericalExpressionTree(
                    construct_expression_tree(effect_node, domain_functions))
                new_action.numeric_effects.add(numerical_precondition)
                continue

    def parse_action(self, action_ast: List[Union[str, List[str]]], domain_types: Dict[str, PDDLType],
                     domain_functions: Dict[str, PDDLFunction],
                     domain_predicates: Dict[str, Predicate],
                     domain_constants: Dict[str, PDDLConstant] = {}) -> Action:
        """Parse a single action AST and returns the object that represent a single action.

        :param action_ast: the AST representation of the action.
        :param domain_types: the types that were extracted from the domain.
        :param domain_functions: the functions that were extracted from the AST.
        :param domain_predicates: the predicates that were extracted from the domain's AST.
        :param domain_constants: the constants that might exist in the domain and might appear in the
            preconditions / effects.
        :return: the action object that is generated using the data from the domain AST.
        """
        self.logger.info("Starting to parse a new action!")
        new_action = Action()
        new_action.name = action_ast[0].lower()
        if len(action_ast[1:]) != 6:  # the number of different sections for the action.
            raise SyntaxError(f"Received an Illegal action AST definition! The action given - {action_ast}")

        action_section_iterator = iter(action_ast[1:])
        for action_label_item in action_section_iterator:
            if action_label_item == ":parameters":
                self.logger.debug(f"Parsing the parameters of the action - {new_action.name}")
                parameters_list = next(action_section_iterator)
                new_action.signature = parse_signature(iter(parameters_list), domain_types)
                continue

            if action_label_item == ":precondition" and not self.partial_parsing:
                self.logger.debug(f"Starting to parse the preconditions of the action - {new_action.name}")
                self.parse_preconditions(next(action_section_iterator), new_action, domain_functions,
                                         domain_predicates, domain_constants)
                continue

            if action_label_item == ":effect" and not self.partial_parsing:
                self.logger.debug(f"Starting to parse the effects of the action - {new_action.name}")
                self.parse_effects(next(action_section_iterator), new_action, domain_functions,
                                   domain_predicates, domain_constants)

        return new_action

    def parse_domain(self) -> Domain:
        """The main entry point that parses the domain file and returns the resulting Domain object.

        :return: the domain object extracted from the tokens.
        """
        domain_expressions = self.tokenizer.parse()
        if domain_expressions[0] != "define":
            raise SyntaxError("Encountered a PDDL that does not start with 'define' statement!")

        domain = Domain()
        for expression in domain_expressions[1:]:
            if expression[0] == "domain":
                domain.name = expression[1]

            elif expression[0] == ":requirements":
                domain.requirements = expression[1:]

            elif expression[0] == ":types":
                domain.types = self.parse_types(expression[1:])

            elif expression[0] == ":constants":
                domain.constants = self.parse_constants(expression[1:], domain.types)

            elif expression[0] == ":predicates":
                domain.predicates = self.parse_predicates(expression[1:], domain.types)

            elif expression[0] == ":functions":
                domain.functions = self.parse_functions(expression[1:], domain.types)

            elif expression[0] == ":action":
                new_action: Action = self.parse_action(expression[1:], domain.types, domain.functions,
                                                       domain.predicates, domain.constants)
                domain.actions[new_action.name] = new_action

            elif expression[0] == ":process" or expression[0] == ":event":
                self.logger.debug("Still no support for temporal domains.")
                # TODO: complete once I finish numeric actions support

        return domain



"""Module containing a parser that is able to parse PDDL domains."""
import logging
from collections import Counter
from pathlib import Path
from typing import NoReturn, List, Union, Dict

from pddl_plus_parser.lisp_parsers import PDDLTokenizer
from pddl_plus_parser.models import Domain, PDDLObject, Problem, PDDLFunction, Predicate, GroundedPredicate, \
    NumericalExpressionTree, construct_expression_tree

LEGAL_GOAL_OPERATORS = [">", "=", "<", ">=", "<="]


class ProblemParser:
    """Class that parses PDDL+ problem files."""

    tokenizer: PDDLTokenizer
    logger: logging.Logger
    domain: Domain
    problem: Problem

    def __init__(self, problem_path: Path, domain: Domain):
        self.tokenizer = PDDLTokenizer(problem_path)
        self.domain = domain
        self.logger = logging.getLogger(__name__)
        self.problem = Problem(domain)

    def _validate_object_types(self, lifted_predicate: Predicate, predicate_signature_items: List[str]) -> NoReturn:
        """Validate that the grounded predicate objects match the lifted predicate signature's type.

        :param lifted_predicate: the lifted predicate that is being compared to.
        :param predicate_signature_items: the objects of the grounded predicate AST.
        """
        optional_objects = {name: pddl_object for name, pddl_object in self.problem.objects.items()}
        optional_objects.update(self.domain.constants)
        objects_types = [optional_objects[obj_name].type for obj_name in predicate_signature_items]
        lifted_predicate_types = list(lifted_predicate.signature.values())
        for index, grounded_object_type in enumerate(objects_types):
            assert grounded_object_type.is_sub_type(lifted_predicate_types[index])

    def parse_domain_name(self, domain_name: str) -> NoReturn:
        """Parse the domain name in the problem file and verifies that the name given matches the inner domain.

        :param domain_name: the name of the domain given as input.
        """
        self.logger.info("Parsing the name of the domain given in the problem definition.")
        if domain_name != self.domain.name:
            raise ValueError(f"Given a problem with mismatching domain. "
                             f"Given domain name - {domain_name} problem domain - {self.domain.name}")

    def parse_objects(self, objects_ast: List[str]) -> Dict[str, PDDLObject]:
        """Parses the objects in the problem definition.

        :param objects_ast: the list of strings representing the typed objects in the problem.
        :return: the list of PDDL objects that are defined in the PDDL problem.
        """
        self.logger.info("Starting to parse the problem's objects.")
        problem_objects = {}
        same_type_objects = []
        iterator = 0
        while iterator < len(objects_ast):
            if isinstance(objects_ast[iterator], list):
                self.logger.debug("Found private objects, iterating over them and then continue regularly.")
                private_objects = self.parse_objects(objects_ast[iterator][1:])
                problem_objects.update(private_objects)
                iterator += 1
                continue

            if objects_ast[iterator] != "-":
                same_type_objects.append(objects_ast[iterator])
                iterator += 1
                continue

            objects_type = objects_ast[iterator + 1]
            if objects_type not in self.domain.types:
                raise ValueError(f"Received illegal objects type - {objects_type}")

            problem_objects.update(
                {name: PDDLObject(name=name, type=self.domain.types[objects_type]) for name in same_type_objects})
            same_type_objects = []
            iterator += 2
            continue

        return problem_objects

    def parse_grounded_numeric_fluent(self, grounded_numeric_fluent: List[str]) -> PDDLFunction:
        """Parse a single grounded numeric fluent in the problem.

        :param grounded_numeric_fluent: the grounded fluent that is present in the problem.
        :return: the function object representing the grounded fluent.
        """
        function_name = grounded_numeric_fluent[0]
        self.logger.info(f"Starting to parse the grounded numeric fluent - {function_name}")
        assert function_name in self.domain.functions
        lifted_function = self.domain.functions[function_name]
        # For now, assuming that fluents have valid parameters.
        fluent_signature_items = grounded_numeric_fluent[1:]
        grounded_fluents_counter = Counter(fluent_signature_items)
        if len(fluent_signature_items) != len(lifted_function.signature):
            raise ValueError(f"Received fluent - {function_name} with wrong number of parameters! "
                             f"Expected - {len(lifted_function.signature)} and received - {len(fluent_signature_items)}")

        possible_objects = {**self.problem.objects, **self.domain.constants}
        fluent_signature = {
            object_name: possible_objects[object_name].type for object_name in fluent_signature_items
        }
        repeating_items = {object_name: count for object_name, count in grounded_fluents_counter.items() if count > 1}
        for grounded_signature_type, lifted_signature_type in zip(fluent_signature.values(),
                                                                  lifted_function.signature.values()):
            assert grounded_signature_type.is_sub_type(lifted_signature_type)

        return PDDLFunction(name=function_name, signature=fluent_signature, repeating_variables=repeating_items)

    def parse_grounded_predicate(self, grounded_predicate_ast: List[str],
                                 lifted_predicate: Predicate) -> GroundedPredicate:
        """Parse the grounded predicate that appears in the problem definition.

        :param grounded_predicate_ast: the AST that represents the grounded predicate.
        :param lifted_predicate: the lifted predicate that the domain defines.
        :return: the grounded predicate represented by the lifted predicate.
        """
        predicate_name = lifted_predicate.name
        predicate_signature_items = grounded_predicate_ast[1:]
        self.logger.info(f"Starting the parse the grounded predicate - {predicate_name} "
                         f"with the signature - {predicate_signature_items}.")
        if len(predicate_signature_items) != len(lifted_predicate.signature):
            raise ValueError(
                f"Received illegal grounded predicate with mismatching signature - {grounded_predicate_ast}")

        self._validate_object_types(lifted_predicate, predicate_signature_items)
        object_mapping = {
            parameter_name: object_name for object_name, parameter_name in
            zip(predicate_signature_items, lifted_predicate.signature)
        }

        return GroundedPredicate(name=predicate_name, signature=lifted_predicate.signature,
                                 object_mapping=object_mapping)

    def parse_state_component(self, expression: List[Union[str, List[str]]]) -> NoReturn:
        """Parse a single AST component that represents either a predicate or a numeric fluent.

        :param expression: the ast component that can represent either a predicate or a numeric fluent.
        """
        self.logger.info("Parsing a single component from the initial state data.")
        if expression[0] == "=":  # This is an assignment of a grounded numeric fluent.
            if len(expression) != 3:  # ['=', <function items as a list>, '<value>']
                raise SyntaxError("A numeric fluent should be of length 3. Fluent scheme: "
                                  "(= (<fluent_name> <argument>) <value>)"
                                  f"Received - {expression}")

            self.logger.debug("Component found is a numeric fluent, starting to process the fluent.")
            function_data = expression[1]
            assigned_value = float(expression[2])
            numeric_fluent = self.parse_grounded_numeric_fluent(function_data)
            self.logger.debug(f"Setting the fluent's value to - {assigned_value}")
            numeric_fluent.set_value(assigned_value)
            self.problem.initial_state_fluents[numeric_fluent.untyped_representation] = numeric_fluent
            return

        if expression[0] in self.domain.predicates:
            self.logger.debug("Component found is a predicate, starting to process it.")
            lifted_predicate = self.domain.predicates[expression[0]]
            grounded_predicate = self.parse_grounded_predicate(expression, lifted_predicate)
            self.problem.initial_state_predicates[lifted_predicate.untyped_representation].add(grounded_predicate)
            return

        raise ValueError(f"Received illegal state component - {expression}")

    def parse_initial_state(self, init_ast: List[List[Union[str, List[str]]]]) -> NoReturn:
        """Parse the initial state of the problem.

        :param init_ast: the AST representation of the inital state.
        """
        self.logger.info("Starting to parse the initial state!")
        for expression in init_ast:
            self.parse_state_component(expression)

    def parse_goal_state(self, goal_state_ast: List[List[Union[str, List[str]]]]) -> NoReturn:
        """Parse the goal state of the problem and extracts the list of grounded predicates from it.

        :param goal_state_ast: the AST representation of the goal state.
        """
        if goal_state_ast[0] != "and":
            raise SyntaxError("Goal state should always start with an AND statement!")

        for expression in goal_state_ast[1:]:
            if expression[0] not in self.domain.predicates and expression[0] not in LEGAL_GOAL_OPERATORS:
                raise ValueError(f"Received illegal state component - {expression}")

            if expression[0] not in LEGAL_GOAL_OPERATORS:  # This is an assignment of a grounded numeric fluent.
                grounded_predicate = self.parse_grounded_predicate(expression, self.domain.predicates[expression[0]])
                self.problem.goal_state_predicates.append(grounded_predicate)
                continue

            numeric_goal_statement = NumericalExpressionTree(
                construct_expression_tree(expression, self.domain.functions))
            self.problem.goal_state_fluents.add(numeric_goal_statement)

    def parse_problem(self) -> Problem:
        """Parse the problem's AST and extracts the object that is represented by the input scheme.

        :return: the problem object.
        """
        problem_expression = self.tokenizer.parse()
        if problem_expression[0] != "define":
            raise SyntaxError("Encountered a PDDL that does not start with 'define' statement!")

        for macro_expression in problem_expression:
            if macro_expression[0] == "problem":
                self.problem.name = macro_expression[1]
                continue

            elif macro_expression[0] == ":domain":
                self.parse_domain_name(macro_expression[1])
                continue

            elif macro_expression[0] == ":objects":
                self.problem.objects = self.parse_objects(macro_expression[1:])

            elif macro_expression[0] == ":init":
                self.parse_initial_state(macro_expression[1:])
                continue

            elif macro_expression[0] == ":goal":
                self.parse_goal_state(macro_expression[1])
                continue

            elif macro_expression == ":metric":
                self.logger.debug("Currently not parsing the metric section of the problem.")
                continue

        return self.problem



#!/usr/bin/python3
import argparse
import itertools
import random
from pathlib import Path

from pddl_plus_parser.problem_generators import get_problem_template

TEMPLATE_FILE_PATH = Path("sailing_template.pddl")


def generate_instance(instance_name: str, num_boats: int, num_people: int, max_dist_goals: int) -> str:
    template = get_problem_template(TEMPLATE_FILE_PATH)
    template_mapping = {
        "instance_name": instance_name,
        "domain_name": "sailing",
        "boat_name_list": " ".join([f"b{i}" for i in range(num_boats)]),
        "boat_positions": "\n".join(
            [f"(= (x b{i}) {random.randint(-10, +10)})\n(= (y b{i}) 0)" for i in range(num_boats)]),
        "people_name_list": " ".join([f"p{i}" for i in range(num_people)]),
        "people_d_position": "\n".join(
            [f"(= (d p{i}) {random.randint(0, max_dist_goals)})" for i in range(num_people)]),
        "people_to_save": "\n".join([f"(saved p{i})" for i in range(num_people)])
    }

    return template.substitute(template_mapping)


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Generate sailing planning instance")
    parser.add_argument("--min_boats", required=True, help="Starting numer of boats")
    parser.add_argument("--max_boats", required=True, help="Maximal number of boats")
    parser.add_argument("--min_people", required=True, help="Min of people to save")
    parser.add_argument("--max_people", required=True, help="Max of people to save")
    parser.add_argument("--max_dist_goal", required=False, help="Max distance people to be rescued", default=500)
    parser.add_argument("--output_folder", required=True, help="Path to the directory containing the generated problems")

    args = parser.parse_args()
    return args


def generate_multiple_problems(
        min_boats: int, max_boats: int, min_people: int, max_people: int, max_dist_goal: int, output_folder: Path):
    boats_range = [i for i in range(min_boats, max_boats + 1)]
    people_range = [i for i in range(min_people, max_people + 1)]
    for num_boats, num_people in itertools.product(boats_range, people_range):
        with open(output_folder / f"pfile{num_boats}_{num_people}_{max_dist_goal}.pddl", "wt") as problem_file:
            problem_file.write(
                generate_instance(f"instance_{num_boats}_{num_people}", num_boats, num_people, max_dist_goal))


def main():
    args = parse_arguments()
    generate_multiple_problems(args.min_boats, args.max_boats, args.min_people, args.max_people, args.max_dist_goal,
                               Path(args.output_folder))


if __name__ == '__main__':
    main()



"""Module to contain all different types of action calls."""
from typing import List

NOP_ACTION = "nop"


class ActionCall:
    """An object representing a single action call."""
    name: str
    parameters: List[str]

    def __init__(self, name: str, grounded_parameters: List[str]):
        self.name = name
        self.parameters = grounded_parameters

    def __str__(self):
        called_objects = " ".join(self.parameters)
        return f"({self.name} {called_objects})"


class JointActionCall:
    """An object representing a single action call."""

    actions: List[ActionCall]

    def __init__(self, actions: List[ActionCall]):
        self.actions = actions

    def __str__(self) -> str:
        return f"[{','.join([str(action) for action in self.actions])}]"

    @property
    def action_count(self) -> int:
        return len([action for action in self.actions if action.name != NOP_ACTION])

    @property
    def joint_parameters(self) -> List[str]:
        params = []
        for action in self.actions:
            params.extend(action.parameters)

        return params



import os
from pathlib import Path

CWD = os.getcwd()
TEST_NUMERIC_PROBLEM = Path(CWD, "test_agricola_problem.pddl")
TEST_NUMERIC_DOMAIN = Path(CWD, "test_agricola_domain.pddl")
TEST_HARD_NUMERIC_DOMAIN = Path(CWD, "satellite_numeric_domain.pddl")




import os
from pathlib import Path

CWD = os.getcwd()
TEST_PARSING_FILE_PATH = Path(CWD, "test_domain_format_file.pddl")
TEST_WOODWORKING_DOMAIN_PATH = Path(CWD, "woodworking_domain.pddl")
TEST_NUMERIC_DEPOT_DOMAIN_PATH = Path(CWD, "depot_numeric_domain.pddl")
TEST_NUMERIC_PROBLEM = Path(CWD, "test_agricola_problem.pddl")
TEST_NUMERIC_DOMAIN = Path(CWD, "test_agricola_domain.pddl")

TEST_NUMERIC_DEPOT_DOMAIN = Path(CWD, "depot_numeric.pddl")
TEST_NUMERIC_DEPOT_PROBLEM = Path(CWD, "pfile2.pddl")
TEST_NUMERIC_DEPOT_TRAJECTORY = Path(CWD, "test_numeric_trajectory")

FARMLAND_NUMERIC_DOMAIN = Path(CWD, "farmland.pddl")
FARMLAND_NUMERIC_PROBLEM = Path(CWD, "pfile10_10.pddl")
FARMLAND_NUMERIC_TRAJECTORY = Path(CWD, "pfile10_10.trajectory")

ZENOTRAVEL_DOMAIN_PATH = Path(CWD, "zenonumeric.pddl")
ZENOTRAVEL_PROBLEM_PATH = Path(CWD, "pfile0.pddl")

PLANT_WATERING_DOMAIN = Path(CWD, "plant_watering_domain.pddl")

WOODWORKING_COMBINED_DOMAIN_PATH = Path(CWD) / "woodworking_combined_domain.pddl"
WOODWORKING_COMBINED_PROBLEM_PATH = Path(CWD) / "woodworking_combined_problem.pddl"
WOODWORKING_COMBINED_TRAJECTORY_PATH = Path(CWD) / "ma_woodworking_trajectory.trajectory"



from typing import List

from pddl_plus_parser.models import Problem, State, ActionCall, Domain, Operator


def create_initial_state(problem: Problem) -> State:
    """Create the initial state of the problem.

    :param problem: the problem object.
    :return: the initial state of the problem.
    """
    initial_state_predicates = problem.initial_state_predicates
    initial_state_numeric_fluents = problem.initial_state_fluents
    return State(predicates=initial_state_predicates, fluents=initial_state_numeric_fluents, is_init=True)


def apply_actions(domain: Domain, current_state: State, joint_action: List[ActionCall]) -> State:
    """

    :param domain: the domain with the action scheme.
    :param current_state: the current state that the action is being applied on.
    :param joint_action: the executable actions of the agents.
    :return:
    """
    interm_state_predicates = current_state.state_predicates.copy()
    interm_state_numeric_fluents = current_state.state_fluents.copy()
    operators = []
    for action_call in joint_action:
        action = domain.actions[action_call.name]
        operator = Operator(action=action, domain=domain, grounded_action_call=action_call.parameters)
        partial_numeric_state = State(predicates=interm_state_predicates, fluents=interm_state_numeric_fluents)
        partial_next_state = operator.apply(partial_numeric_state)
        interm_state_predicates.update(partial_next_state.state_predicates)
        interm_state_numeric_fluents.update(partial_next_state.state_fluents)
        operators.append(operator)

    return State(predicates=interm_state_predicates, fluents=interm_state_numeric_fluents)



"""Module representing the objects and constants of PDDL files."""
from .pddl_type import PDDLType


class PDDLObject:
    name: str
    type: PDDLType

    def __init__(self, name: str, type: PDDLType):
        self.name = name
        self.type = type

    def __str__(self):
        return f"{self.name} - {self.type.name}"


PDDLConstant = PDDLObject



import os
from pathlib import Path

CWD = os.getcwd()
SOKOBAN_DOMAIN_FILE_PATH = Path(CWD, "sokoban_domain.pddl")
SOKOBAN_UNPARSED_PLAN_PATH = Path(CWD, "sokoban_plan.txt")
SOKOBAN_PROBLEM_PATH = Path(CWD, "sokoban_problem.pddl")

WOODWORKING_DOMAIN_FILE_PATH = Path(CWD, "woodworking_domain.pddl")
WOODWORKING_UNPARSED_PLAN_PATH = Path(CWD, "woodworking_plan.txt")
WOODWORKING_PARSED_PLAN_PATH = Path(CWD, "woodworking_plan.solution")
WOODWORKING_SHORT_PARSED_PLAN_PATH = Path(CWD, "woodworking_short_plan.solution")

COMBINED_PROBLEM_PATH = Path(CWD, "combined_problem.pddl")
COMBINED_DOMAIN_PATH = Path(CWD, "combined_domain.pddl")

MULTI_AGENT_DATA_DIRECTORY = Path(CWD, "multi_agent_problem")

WOODWORKING_AGENT_NAMES = ["glazer0", "grinder0", "highspeed-saw0", "immersion-varnisher0", "planer0", "saw0",
                           "spray-varnisher0"]



"""Module that represents a numerical in a PDDL+ model."""
from typing import Optional, NoReturn, Dict

from .pddl_predicate import SignatureType


class PDDLFunction:
    """Class that represents a numerical function."""

    name: str
    signature: SignatureType
    stored_value: float
    repeating_variables: Dict[str, int]

    def __init__(self, name: Optional[str] = None, signature: Optional[SignatureType] = None,
                 repeating_variables: Optional[Dict[str, int]] = {}):
        self.name = name
        self.signature = signature
        self.stored_value = 0
        self.repeating_variables = repeating_variables

    def __eq__(self, other: "PDDLFunction") -> bool:
        """Checks whether two functions are considered equal.

        Equality can be considered if a type inherits from another type as well.

        :param other: the other predicate to compare.
        :return: whether the predicates are equal.
        """
        if not self.name == other.name:
            return False

        for parameter_name, parameter_type in self.signature.items():
            if parameter_name not in other.signature:
                return False

            other_param_type = other.signature[parameter_name]
            if not parameter_type.is_sub_type(other_param_type):
                return False

        return True

    @property
    def value(self) -> float:
        return self.stored_value

    def set_value(self, value: float) -> NoReturn:
        """Set the value of a function to be the input value.

        :param value: the value to set to the function.
        """
        self.stored_value = value

    @property
    def state_representation(self) -> str:
        """Returns the state representation of the function."""
        function_variables = []
        for repeating_variable, num_repeats in self.repeating_variables.items():
            function_variables.extend([repeating_variable] * num_repeats)
        function_variables.extend(param for param in self.signature if param not in self.repeating_variables)

        untyped_signature_str = " ".join(function_variables)
        return f"(= ({self.name} {untyped_signature_str}) {self.value})"

    @property
    def state_typed_representation(self) -> str:
        """Returns the state representation of the function with the type signature."""
        function_variables = []
        for repeating_variable, num_repeats in self.repeating_variables.items():
            function_variables.extend([repeating_variable] * num_repeats)
        function_variables.extend(param for param in self.signature if param not in self.repeating_variables)

        signature_str_items = [f"{parameter_name} - {str(self.signature[parameter_name])}"
                               for parameter_name in function_variables]
        return f"(= ({self.name} {' '.join(signature_str_items)}) {self.value})"

    @property
    def untyped_representation(self) -> str:
        """Returns the representation of the function without the type information.

        Note:
            This property is used only for lifted assignments so no need to check for the grounded items' multiplicity.
        """
        untyped_signature_str = " ".join(self.signature.keys())
        return f"({self.name} {untyped_signature_str})"

    def __str__(self):
        signature_str_items = []
        for parameter_name, parameter_type in self.signature.items():
            signature_str_items.append(f"{parameter_name} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({self.name} {signature_str})"



import string
from pathlib import Path


def get_problem_template(template_file_path: Path) -> string.Template:
    """

    :return:
    """
    with open(template_file_path, "rt") as template_file:
        text = template_file.read()
        return string.Template(text)



from distutils.core import setup
from setuptools import find_packages

with open("README.md") as f:
    long_description = f.read()

setup(name='pddl-plus-parser',
      version='2.2.1',
      python_requires=">=3.8",
      description='Parser of PDDL+ domains and problems for learning purposes',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Argaman Mordoch',
      packages=find_packages(exclude=["tests"]),
     )


"""Module that contains the definition of PDDL+ types"""
from typing import Optional


class PDDLType:
    name: str
    parent: Optional["PDDLType"]

    def __init__(self, name: str, parent: Optional["PDDLType"] = None):
        self.name = name.lower()
        self.parent = parent

    def __str__(self):
        return self.name

    def __repr__(self):
        if self.parent is not None:
            return f"type: {self.name} descendant of {str(self.parent)}"

        return f"type: {self.name}"

    def __eq__(self, other: "PDDLType"):
        return self.name == other.name

    def is_sub_type(self, other_type: "PDDLType") -> bool:
        """Checks if a type a subtype of the other.

        :param other_type: the type that is checked to se if the first is subtype of.
        :return: whether or not the first is a subtype of the other.
        """
        compared_type = self
        while compared_type.parent is not None:
            if compared_type.name == other_type.name:
                return True

            compared_type = compared_type.parent

        return False



"""This module exports the domain object to a domain file."""
from collections import defaultdict
from pathlib import Path
from typing import NoReturn, List, Dict, Set

from pddl_plus_parser.models import Predicate, Action, Domain, PDDLConstant, PDDLFunction, PDDLType


class DomainExporter:
    """Class that is able to export a domain to a correct PDDL file."""

    @staticmethod
    def write_action_predicates(positive_predicates: Set[Predicate], negative_predicates: Set[Predicate]) -> str:
        """Write the effects of an action according to the PDDL file format.

        :param positive_predicates: the add effects of an action.
        :param negative_predicates: the delete effects of an action.
        :return: the formatted string representing the action's predicates (preconditions or effects).
        """
        action_effect_predicates = [predicate.untyped_representation for predicate in positive_predicates]
        action_effect_predicates.extend(
            [f"(not {predicate.untyped_representation})" for predicate in negative_predicates])
        formatted_effects = "{content}"
        return formatted_effects.format(content=" ".join(action_effect_predicates))

    def write_action(self, action: Action) -> str:
        """Write the action formatted string from the action data.

        :param action: The action that needs to be formatted into a string.
        :return: the string format of the action.
        """
        action_params = " ".join(
            [f"{name} - {parameter_type.name}" for name, parameter_type in action.signature.items()])

        discrete_preconditions = self.write_action_predicates(action.positive_preconditions,
                                                              action.negative_preconditions)
        discrete_preconditions = f"(and {discrete_preconditions})"
        numeric_preconditions = "\n\t\t".join([fluent.to_pddl() for fluent in action.numeric_preconditions])
        if len(numeric_preconditions) > 0:
            numeric_preconditions += "\n"

        discrete_effects = self.write_action_predicates(action.add_effects, action.delete_effects)
        numeric_effects = "\n\t".join([fluent.to_pddl() for fluent in action.numeric_effects])
        if len(numeric_effects) > 0:
            numeric_effects += "\n"

        return f"(:action {action.name}\n" \
               f"\t:parameters   ({action_params})\n" \
               f"\t:precondition {discrete_preconditions}\n{numeric_preconditions}" \
               f"\t:effect       (and {discrete_effects}\n{numeric_effects})" \
               f")\n"

    @staticmethod
    def write_predicates(predicates: Dict[str, Predicate]) -> str:
        """Writes the predicates formatted according to the domain file format.

        :param predicates: the predicates that are in the domain's definition.
        :return: the formatted string representing the domain's predicates.
        """
        predicates_str = "(:predicates\n{predicates})\n\n"
        predicates_strings = []
        for predicate_name, predicate in predicates.items():
            predicate_params = " ".join(
                [f"{name} - {types[0]}" for name, types in predicate.signature])
            predicates_strings.append(f"\t({predicate_name} {predicate_params})")

        return predicates_str.format(predicates="\n".join(predicates_strings))

    @staticmethod
    def write_constants(constants: Dict[str, PDDLConstant]) -> str:
        """Writes the constants of the domain to the new domain file.

        :param constants: the constants that appear in the domain object.
        :return: the representation of the constants as a canonical PDDL string.
        """
        same_type_constant = defaultdict(list)
        for const_name, constant in constants.items():
            if const_name == "object":
                continue

            same_type_constant[constant.type.name].append(const_name)

        types_strs = []
        for constant_type_name, constant_objects in same_type_constant.items():
            types_strs.append(f"\t{' '.join(constant_objects)} - {constant_type_name}")

        return "\n".join(types_strs)

    @staticmethod
    def format_type_like_string(sorted_type_like_objects: Dict[str, List[str]]) -> List[str]:
        """formats the string that are of the same format as types. This applies to both consts and types.

        :param sorted_type_like_objects: the type like objects that are being formatted into a list of strings.
        :return: the formatted strings as a list.
        """
        type_like_object_content = []
        for pddl_type_name, sub_types in sorted_type_like_objects.items():
            type_like_object_pddl_str = "\t"
            type_like_object_pddl_str += "\n\t".join([child_type for child_type in sub_types[:-1]])
            type_like_object_pddl_str += f"\n\t{sub_types[-1]} - {pddl_type_name}"
            type_like_object_content.append(type_like_object_pddl_str)

        return type_like_object_content

    @staticmethod
    def write_types(types: Dict[str, PDDLType]) -> str:
        """Writes the definitions of the types according to the PDDL file format.

        :param types: the types that are available in the learned domain.
        :return: the formatted string representing the types in the PDDL domain file.
        """
        parent_child_map = defaultdict(list)
        for type_name, type_obj in types.items():
            if type_name == "object":
                continue

            parent_child_map[type_obj.parent.name].append(type_name)

        types_strs = []
        for parent_type, children_types in parent_child_map.items():
            types_strs.append(f"\t{' '.join(children_types)} - {parent_type}")

        return "\n".join(types_strs)

    @staticmethod
    def write_functions(functions: Dict[str, PDDLFunction]) -> str:
        """Converts the functions to PDDL format.

        :return: the PDDL format of the functions.
        """
        return "\n\t".join([str(f) for f in functions.values()])

    def extract_domain(self, domain: Domain) -> str:
        """Export the domain object to a correct PDDL file.

        :param domain: the learned domain object.
        """
        predicates = "\n\t".join([str(p) for p in domain.predicates.values()])
        actions = "\n".join(self.write_action(action) for action in domain.actions.values())
        constants = f"(:constants {self.write_constants(domain.constants)}\n)\n\n" if len(domain.constants) > 0 else ""
        functions = f"(:functions {self.write_functions(domain.functions)}\n)\n\n" if len(domain.functions) > 0 else ""
        return f"(define (domain {domain.name})\n" \
               f"(:requirements {' '.join(domain.requirements)})\n" \
               f"(:types {self.write_types(domain.types)}\n)\n\n" \
               f"(:predicates {predicates}\n)\n\n" \
               f"{constants}" \
               f"{functions}" \
               f"{actions}\n)"

    def export_domain(self, domain: Domain, export_path: Path) -> NoReturn:
        """Export the domain object to a correct PDDL file.

        :param domain: the domain object to be exported.
        :param export_path: the path to export the domain to.
        """
        domain_str = self.extract_domain(domain)
        with open(export_path, "wt") as export_domain_file:
            export_domain_file.write(domain_str)



"""Module test for the numeric trajectory exporter functionality."""
import os
from collections import defaultdict
from pathlib import Path
from typing import List

from pytest import fixture

from pddl_plus_parser.exporters import TrajectoryExporter
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser
from pddl_plus_parser.models import Domain, State, Problem

CWD = os.getcwd()
TEST_DISCRETE_DOMAIN_PATH = Path(CWD, "elevators_domain.pddl")
TEST_DISCRETE_PROBLEM_PATH = Path(CWD, "elevators_p03.pddl")
TEST_DISCRETE_PLAN_PATH = Path(CWD, "elevators_p03_plan.solution")
TEST_NUMERIC_DOMAIN_PATH = Path(CWD, "depot_numeric.pddl")
TEST_NUMERIC_PROBLEM_PATH = Path(CWD, "pfile2.pddl")
TEST_NUMERIC_PLAN_PATH = Path(CWD, "depot_numeric.solution")
TEST_DISCRETE_TRAJECTORY_FILE_PATH = Path(CWD, "test_trajectory")
TEST_NUMERIC_TRAJECTORY_FILE_PATH = Path(CWD, "test_numeric_trajectory")


@fixture()
def discrete_domain() -> Domain:
    domain_parser = DomainParser(TEST_DISCRETE_DOMAIN_PATH)
    return domain_parser.parse_domain()


@fixture()
def numeric_domain() -> Domain:
    domain_parser = DomainParser(TEST_NUMERIC_DOMAIN_PATH)
    return domain_parser.parse_domain()


@fixture()
def discrete_problem_parser(discrete_domain: Domain) -> ProblemParser:
    return ProblemParser(problem_path=TEST_DISCRETE_PROBLEM_PATH, domain=discrete_domain)


@fixture()
def numeric_problem_parser(numeric_domain: Domain) -> ProblemParser:
    return ProblemParser(problem_path=TEST_NUMERIC_PROBLEM_PATH, domain=numeric_domain)


@fixture()
def discrete_problem(discrete_problem_parser: ProblemParser) -> Problem:
    return discrete_problem_parser.parse_problem()


@fixture()
def numeric_problem(numeric_problem_parser: ProblemParser) -> Problem:
    return numeric_problem_parser.parse_problem()


@fixture()
def discrete_trajectory_exporter(discrete_domain: Domain) -> TrajectoryExporter:
    return TrajectoryExporter(domain=discrete_domain)


@fixture()
def numeric_trajectory_exporter(numeric_domain: Domain) -> TrajectoryExporter:
    return TrajectoryExporter(domain=numeric_domain)


def create_state_from_predicate_components(
        components: List[List[str]], problem_parser: ProblemParser) -> State:
    problem_parser.parse_problem()
    problem_parser.problem.initial_state_predicates = defaultdict(set)
    problem_parser.problem.initial_state_fluents = {}
    problem_parser.parse_initial_state(components)
    problem = problem_parser.problem
    initial_state_predicates = problem.initial_state_predicates
    initial_state_numeric_fluents = problem.initial_state_fluents
    return State(predicates=initial_state_predicates, fluents=initial_state_numeric_fluents)


def test_create_single_triplet_with_single_state_and_single_call_creates_next_state_correctly(
        discrete_trajectory_exporter: TrajectoryExporter, discrete_problem_parser: ProblemParser):
    test_simple_state_components = [["lift-at", "slow2-0", "n17"], ["above", "n16", "n17"],
                                    ["reachable-floor", "slow2-0", "n16"]]
    test_action_call = "(move-down-slow slow2-0 n17 n16)"
    state = create_state_from_predicate_components(test_simple_state_components, discrete_problem_parser)
    result_triplet = discrete_trajectory_exporter.create_single_triplet(state, test_action_call)
    result_next_state = result_triplet.next_state
    lift_at_predicate = result_next_state.state_predicates["(lift-at ?lift ?floor)"]
    untyped_predicates = [p.untyped_representation for p in lift_at_predicate]
    assert "(lift-at slow2-0 n16)" in untyped_predicates
    assert "(lift-at slow2-0 n17)" not in untyped_predicates


def test_create_single_triplet_with_complete_initial_state_removes_correct_predicates_and_adds_correct_predicates(
        discrete_trajectory_exporter: TrajectoryExporter, discrete_problem: Problem):
    state = State(predicates=discrete_problem.initial_state_predicates, fluents={})
    test_action_call = "(move-down-slow slow2-0 n17 n16)"
    result_triplet = discrete_trajectory_exporter.create_single_triplet(state, test_action_call)
    lift_at_predicate = result_triplet.next_state.state_predicates["(lift-at ?lift ?floor)"]
    untyped_predicates = [p.untyped_representation for p in lift_at_predicate]
    assert "(lift-at slow2-0 n16)" in untyped_predicates
    assert "(lift-at slow2-0 n17)" not in untyped_predicates


def test_create_single_triplet_with_complete_numeric_initial_state_removes_correct_predicates_and_adds_correct_predicates(
        numeric_trajectory_exporter: TrajectoryExporter, numeric_problem: Problem):
    state = State(predicates=numeric_problem.initial_state_predicates, fluents=numeric_problem.initial_state_fluents)
    test_action_call = "(DRIVE TRUCK0 DEPOT0 DISTRIBUTOR0)"
    result_triplet = numeric_trajectory_exporter.create_single_triplet(state, test_action_call)
    test_next_state = result_triplet.next_state
    fuel_level_fluent = test_next_state.state_fluents["(fuel-cost )"]
    assert fuel_level_fluent.value == 10


def test_export_discrete_trajectory(discrete_trajectory_exporter: TrajectoryExporter, discrete_problem: Problem):
    triplets = discrete_trajectory_exporter.parse_plan(discrete_problem, TEST_DISCRETE_PLAN_PATH)
    exportable_triplets = discrete_trajectory_exporter.export(triplets)
    with open(TEST_DISCRETE_TRAJECTORY_FILE_PATH, "wt") as trajectory_file:
        trajectory_file.writelines(exportable_triplets)


def test_export_numeric_trajectory(numeric_trajectory_exporter: TrajectoryExporter, numeric_problem: Problem):
    triplets = numeric_trajectory_exporter.parse_plan(numeric_problem, TEST_NUMERIC_PLAN_PATH)
    exportable_triplets = numeric_trajectory_exporter.export(triplets)
    with open(TEST_NUMERIC_TRAJECTORY_FILE_PATH, "wt") as trajectory_file:
        trajectory_file.writelines(exportable_triplets)



"""Parse the output of the ENHSP planner so that the algorithms will be able to use it."""
import logging
import sys

from pathlib import Path
from typing import NoReturn, List


class ENHSPParser:
    """Parse metricFF plans and exports then into standard output file."""

    logger: logging.Logger

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @staticmethod
    def parse_plan_content(input_path: Path) -> List[str]:
        """Parse the content of the file and export the action sequence.

        :param input_path: the path to the plan sequence file generated by the solver.
        :return: the action sequence.
        """
        plan_seq = []
        with open(input_path, "rt") as plan_file:
            for line in plan_file.readlines():
                plan_seq.append(line.lower())

        return plan_seq

    def parse_plan(self, input_path: Path) -> NoReturn:
        """Parse the output file and exports a plan if exists.

        :param input_path: the path to the output log of metricFF planner.
        """
        action_sequence = self.parse_plan_content(input_path)
        with open(input_path, "wt") as output_file:
            output_file.writelines(action_sequence)


if __name__ == '__main__':
    ENHSPParser().parse_plan(Path(sys.argv[1]))
